#include "driver/i2s_reg.h"
#include "driver/slc_register.h"
#include "driver/sdio_slv.h"
#include "driver/i2s.h"
#include "osapi.h"
#include "gpio.h"
#include "pin_mux_register.h"
#include "jbuf.h"
#include "pwm.h"

#define CONF_RXLINK_ADDR(addr)		CLEAR_PERI_REG_MASK(SLC_RX_LINK,SLC_RXLINK_DESCADDR_MASK);\
	SET_PERI_REG_MASK(SLC_RX_LINK, ((uint32)(addr)) & SLC_RXLINK_DESCADDR_MASK)
#define CONF_TXLINK_ADDR(addr) 		CLEAR_PERI_REG_MASK(SLC_TX_LINK,SLC_TXLINK_DESCADDR_MASK);\
	SET_PERI_REG_MASK(SLC_TX_LINK, ((uint32)(addr)) & SLC_TXLINK_DESCADDR_MASK)

#define START_RXLINK() 	SET_PERI_REG_MASK(SLC_RX_LINK, SLC_RXLINK_START)
#define START_TXLINK() 	SET_PERI_REG_MASK(SLC_TX_LINK, SLC_TXLINK_START)

#define i2c_bbpll                                 0x67
#define i2c_bbpll_en_audio_clock_out            4
#define i2c_bbpll_en_audio_clock_out_msb        7
#define i2c_bbpll_en_audio_clock_out_lsb        7
#define i2c_bbpll_hostid                           4

#define i2c_writeReg_Mask(block, host_id, reg_add, Msb, Lsb, indata)  rom_i2c_writeReg_Mask(block, host_id, reg_add, Msb, Lsb, indata)
#define i2c_readReg_Mask(block, host_id, reg_add, Msb, Lsb)  rom_i2c_readReg_Mask(block, host_id, reg_add, Msb, Lsb)
#define i2c_writeReg_Mask_def(block, reg_add, indata) \
      i2c_writeReg_Mask(block, block##_hostid,  reg_add,  reg_add##_msb,  reg_add##_lsb,  indata)
#define i2c_readReg_Mask_def(block, reg_add) \
      i2c_readReg_Mask(block, block##_hostid,  reg_add,  reg_add##_msb,  reg_add##_lsb)

#define IIS_RX_BUF_LEN  2560  //unit Byte

#define RX_NUM (IIS_RX_BUF_LEN/4) //unit word

uint32 i2s_rx_buff1[IIS_RX_BUF_LEN/4] = {0};
uint32 i2s_rx_buff2[IIS_RX_BUF_LEN/4] = {0};

struct sdio_queue i2s_rx_queue1, i2s_rx_queue2;

void slc_isr(void *para);

//create DMA buffer descriptors, unit of either size or length here is byte.
//More details in I2S documents.
void creat_one_link(uint8 own, uint8 eof,uint8 sub_sof, uint16 size, uint16 length,
                    uint32* buf_ptr, struct sdio_queue* nxt_ptr, struct sdio_queue* i2s_queue) {
    i2s_queue->owner=own;
    i2s_queue->eof= eof;
    i2s_queue->sub_sof=sub_sof;
    i2s_queue->datalen=length;
    i2s_queue->blocksize=size;
    i2s_queue->buf_ptr=(uint32)buf_ptr;
    i2s_queue->next_link_ptr=(uint32)nxt_ptr;
    i2s_queue->unused=0;
}

#define BASEFREQ (160000000L)
#define ABS(x) (((x)>0)?(x):(-(x)))

//Set the I2S sample rate, in HZ
void i2sSetRate(int rate, int enaWordlenFuzzing) {
	//Find closest divider
	int bestclkmdiv, bestbckdiv, bestbits, bestfreq=-10000;
	int tstfreq;
	int bckdiv, clkmdiv, bits=16;
	/*
		CLK_I2S = 160MHz / I2S_CLKM_DIV_NUM
		BCLK = CLK_I2S / I2S_BCK_DIV_NUM
		WS = BCLK/ 2 / (16 + I2S_BITS_MOD)
		Note that I2S_CLKM_DIV_NUM must be >5 for I2S data
		I2S_CLKM_DIV_NUM - 5-127
		I2S_BCK_DIV_NUM - 2-127

		We also have the option to send out more than 2x16 bit per sample. Most I2S codecs will
		ignore the extra bits and in the case of the 'fake' PWM/delta-sigma outputs, they will just lower the output
		voltage a bit, so we add them when it makes sense.
	*/
	for (bckdiv=2; bckdiv<64; bckdiv++) {
		for (clkmdiv=5; clkmdiv<64; clkmdiv++) {
			for (bits=16; bits<(enaWordlenFuzzing?20:17); bits++) {
				tstfreq=BASEFREQ/(bckdiv*clkmdiv*bits*2);
				if (ABS(rate-tstfreq)<ABS(rate-bestfreq)) {
					bestfreq=tstfreq;
					bestclkmdiv=clkmdiv;
					bestbckdiv=bckdiv;
					bestbits=bits;
				}
			}
		}
	}

	ets_uart_printf("ReqRate %d MDiv %d BckDiv %d Bits %d  Frq %d\n",
		rate, bestclkmdiv, bestbckdiv, bestbits, (int)(BASEFREQ/(bestbckdiv*bestclkmdiv*bestbits*2)));


	CLEAR_PERI_REG_MASK(I2SCONF, I2S_TRANS_SLAVE_MOD|
						(I2S_BITS_MOD<<I2S_BITS_MOD_S)|
						(I2S_BCK_DIV_NUM <<I2S_BCK_DIV_NUM_S)|
						(I2S_CLKM_DIV_NUM<<I2S_CLKM_DIV_NUM_S));
	SET_PERI_REG_MASK(I2SCONF, I2S_RIGHT_FIRST|I2S_MSB_RIGHT|I2S_RECE_SLAVE_MOD|
						I2S_RECE_MSB_SHIFT|I2S_TRANS_MSB_SHIFT|
						((bestbits-16)<<I2S_BITS_MOD_S)|
						(((bestbckdiv)&I2S_BCK_DIV_NUM )<<I2S_BCK_DIV_NUM_S)|
						(((bestclkmdiv)&I2S_CLKM_DIV_NUM)<<I2S_CLKM_DIV_NUM_S));
}

void i2s_init() {
    //CONFIG I2S TX PIN FUNC
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_U0RXD_U, FUNC_I2SO_DATA);
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_GPIO2_U, FUNC_I2SO_WS);
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_MTDO_U, FUNC_I2SO_BCK);

    //Enable a 160MHz clock to i2s subsystem
    i2c_writeReg_Mask_def(i2c_bbpll, i2c_bbpll_en_audio_clock_out, 1);
    //reset I2S
    CLEAR_PERI_REG_MASK(I2SCONF,I2S_I2S_RESET_MASK);
    SET_PERI_REG_MASK(I2SCONF,I2S_I2S_RESET_MASK);
    CLEAR_PERI_REG_MASK(I2SCONF,I2S_I2S_RESET_MASK);

    //Enable FIFO in i2s module
    SET_PERI_REG_MASK(I2S_FIFO_CONF, I2S_I2S_DSCR_EN);
    //set I2S_FIFO
#if 0
    //set rx,tx data size, both are "24-bit full data discountinue" here
    SET_PERI_REG_MASK(I2S_FIFO_CONF, (I2S_I2S_RX_FIFO_MOD<<I2S_I2S_RX_FIFO_MOD_S)|(I2S_I2S_TX_FIFO_MOD<<I2S_I2S_TX_FIFO_MOD_S));
#else
	//Select 16bits per channel (FIFO_MOD=0), no DMA access (FIFO only)
	CLEAR_PERI_REG_MASK(I2S_FIFO_CONF, (I2S_I2S_RX_FIFO_MOD<<I2S_I2S_RX_FIFO_MOD_S)|(I2S_I2S_TX_FIFO_MOD<<I2S_I2S_TX_FIFO_MOD_S));
#endif
    //set I2S_CHAN
    //set rx,tx channel mode, both are "two channel" here
    SET_PERI_REG_MASK(I2SCONF_CHAN, (I2S_TX_CHAN_MOD<<I2S_TX_CHAN_MOD_S)|(I2S_RX_CHAN_MOD<<I2S_RX_CHAN_MOD_S));
    //set RX eof num
    //Register I2SRXEOF_NUM sets the number of data to be received when the Rx FIFO triggers the SLC transport (unit: 4 bytes).
    WRITE_PERI_REG(I2SRXEOF_NUM, RX_NUM);

    /*
    in the I2SCONF :
    Bits16~21 are the prescaler of the input clock (I2S_CLKM_DIV_NUM).
    Bits22~27 are the frequency divider of the communication clock signal (I2S_BCK_DIV_NUM).
    */

	//trans master&rece slave,MSB shift,right_first,msb right
	CLEAR_PERI_REG_MASK(I2SCONF, I2S_TRANS_SLAVE_MOD|
						(I2S_BITS_MOD<<I2S_BITS_MOD_S)|
						(I2S_BCK_DIV_NUM <<I2S_BCK_DIV_NUM_S)|
						(I2S_CLKM_DIV_NUM<<I2S_CLKM_DIV_NUM_S));
	SET_PERI_REG_MASK(I2SCONF, I2S_RIGHT_FIRST|I2S_MSB_RIGHT|I2S_RECE_SLAVE_MOD|
						I2S_RECE_MSB_SHIFT|I2S_TRANS_MSB_SHIFT|
						((16&I2S_BCK_DIV_NUM )<<I2S_BCK_DIV_NUM_S)|
						((7&I2S_CLKM_DIV_NUM)<<I2S_CLKM_DIV_NUM_S));
	//clear int
	SET_PERI_REG_MASK(I2SINT_CLR,   I2S_I2S_TX_REMPTY_INT_CLR|I2S_I2S_TX_WFULL_INT_CLR|
			I2S_I2S_RX_WFULL_INT_CLR|I2S_I2S_PUT_DATA_INT_CLR|I2S_I2S_TAKE_DATA_INT_CLR);
	CLEAR_PERI_REG_MASK(I2SINT_CLR,   I2S_I2S_TX_REMPTY_INT_CLR|I2S_I2S_TX_WFULL_INT_CLR|
			I2S_I2S_RX_WFULL_INT_CLR|I2S_I2S_PUT_DATA_INT_CLR|I2S_I2S_TAKE_DATA_INT_CLR);
	//enable int
	SET_PERI_REG_MASK(I2SINT_ENA,   I2S_I2S_TX_REMPTY_INT_ENA|I2S_I2S_TX_WFULL_INT_ENA|
	I2S_I2S_RX_REMPTY_INT_ENA|I2S_I2S_TX_PUT_DATA_INT_ENA|I2S_I2S_RX_TAKE_DATA_INT_ENA);

	//Start transmission
	SET_PERI_REG_MASK(I2SCONF,I2S_I2S_TX_START);

    i2sSetRate(32000, 0);
}

//Initialize the SLC module for DMA function
void slc_init() {
    //Reset DMA
    SET_PERI_REG_MASK(SLC_CONF0, SLC_RXLINK_RST|SLC_TXLINK_RST);
    CLEAR_PERI_REG_MASK(SLC_CONF0, SLC_RXLINK_RST|SLC_TXLINK_RST);

    //Enable and configure DMA
    CLEAR_PERI_REG_MASK(SLC_CONF0, (SLC_MODE<<SLC_MODE_S));
    SET_PERI_REG_MASK(SLC_CONF0, (1<<SLC_MODE_S));
    SET_PERI_REG_MASK(SLC_RX_DSCR_CONF, SLC_INFOR_NO_REPLACE|SLC_TOKEN_NO_REPLACE);//|0xfe
    CLEAR_PERI_REG_MASK(SLC_RX_DSCR_CONF, SLC_RX_FILL_EN|SLC_RX_EOF_MODE | SLC_RX_FILL_MODE);

    ETS_SLC_INTR_ATTACH(slc_isr, NULL);
    /////enable sdio operation intr
    WRITE_PERI_REG(SLC_INT_ENA, SLC_INTEREST_EVENT);
    /////clear sdio initial random active intr signal
    WRITE_PERI_REG(SLC_INT_CLR, 0xffffffff);
    /////enable sdio intr in cpu
    ETS_SLC_INTR_ENABLE();
}


//interrupt
//write new buff data into RX_DMA at the transmitter side
//or load buff data out from TX_DMA at the receiver side
void slc_isr(void *para) {
    int i;
    uint32 slc_intr_status = READ_PERI_REG(SLC_INT_STATUS);
    if (slc_intr_status == 0) {
        //No interested interrupts pending
        return;
    }
    //clear all intrs
    WRITE_PERI_REG(SLC_INT_CLR, 0xffffffff);

    //Transmitter side
    if (slc_intr_status & SLC_RX_EOF_INT_ST) {
        //find the DMA which sends the interrupt signal
        uint32 addr = READ_PERI_REG(SLC_RX_EOF_DES_ADDR);
        short *samples = jbuf_get();
        //ets_uart_printf("S: %04X %04X\n", (unsigned)samples[0], (unsigned)samples[1]);
        if(addr == (uint32)&i2s_rx_queue1) {
            //replace data in the buff
            for (i=0; i<IIS_RX_BUF_LEN/4; i++) {
                i2s_rx_buff1[i] = sampToI2sDeltaSigma(samples[i/4]);
            }
        } else if(addr == (uint32)&i2s_rx_queue2) {
            for (i=0; i<IIS_RX_BUF_LEN/4; i++) {
                i2s_rx_buff2[i] = sampToI2sDeltaSigma(samples[i/4]);
            }
        } else {
            ets_uart_printf("Unexpected!\n");
        }
    }
}


void ICACHE_FLASH_ATTR i2s_run(void) {
    slc_init();

    //To receive data from the I2S module, counter-intuitively we use the TXLINK part, not the RXLINK part and vice versa.
    //Note:At the transimitter side,the size of the DMAs can not be smaller than 128*4 bytes which are the
    //size of the I2S FIFO.
    //Note:At the receiver side,the number of the DMAs can not be smaller than 3 which is limited by the hardware.
    creat_one_link(1, 1, 0, IIS_RX_BUF_LEN, IIS_RX_BUF_LEN, i2s_rx_buff1, &i2s_rx_queue2, &i2s_rx_queue1);
    creat_one_link(1, 1, 0, IIS_RX_BUF_LEN, IIS_RX_BUF_LEN, i2s_rx_buff2, &i2s_rx_queue1, &i2s_rx_queue2);

    //config rx&tx link to hardware
    CONF_RXLINK_ADDR(&i2s_rx_queue1);
    //config rx control, start
    START_RXLINK();

    i2s_init();	//start
}
